// bcell_arith.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// Arithmetic Coding functions (Adaptive)
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

// Based on the ANSI C Arithmetic Coding Library by Fred Wheeler (wheeler@cipr.rpi.edu),
// which was adapted from the program in "Arithmetic Coding for Data Compression", 
// by Ian H. Witten, Radford M. Neal and John G. Cleary, Communications of the ACM,
// June 1987, Vol. 30, No. 6.

#include "bcell.hh"

// Compress our cell using arithmetic coding and return our cell
Bcell& Bcell::arith() {
    const Bcell ullCell((unsigned long long) size);

    ArithEncoder ae;

    for (size_t i = 0; i < size; i++)  {
        ae.encodeSymbol((*this)[i]);
    }

    Bcell bits = ae.finalize();

    *this = ullCell + bits;
    return *this;
}

// Decompress our cell using arithmetic coding and return our cell
Bcell& Bcell::unarith() {
    Bcell uncompressed("unarith() Uncompressed Cell", readULL(0));
    truncate(8, size);

    ArithDecoder ad(this);

    for (size_t i = 0; i < uncompressed.size; i++) {
        uncompressed[i] = ad.decodeSymbol();
    }

    *this = uncompressed;
    return *this;
}

Bcell::ArithEncoder::ArithEncoder() {
    finalized = false;
    low = 0;
    high = TOP_VALUE;
    fbits = 0;
}

void Bcell::ArithEncoder::encodeSymbol(int sym) {
    if (finalized) {
        cout << "Trying to encodeSymbol() with a finalized arithmetic encoder!" << endl;
        exit(EXIT_FAILURE);
    }

    unsigned long int range = high - low + 1;
    high = low + (range * acm[sym]) / acm[0] - 1;
    low  = low + (range * acm[sym + 1]) / acm[0];

    for (;;) {
        if (high < HALF) {
            bitPlusFollow(0);
        } else if (low >= HALF) {
            bitPlusFollow(1);
            low -= HALF;
            high -= HALF;
        } else if (low >= FIRST_QTR && high < THIRD_QTR) {
            fbits++;
            low -= FIRST_QTR;
            high -= FIRST_QTR;
        } else {
            break;
        }

        low = 2 * low;
        high = 2 * high + 1;
    }

    acm.update(sym);
}

void Bcell::ArithEncoder::bitPlusFollow(bool bit) {
    if (finalized) {
        cout << "Trying to bitPlusFollow() with a finalized arithmetic encoder!" << endl;
        exit(EXIT_FAILURE);
    }

    g.push_back(bit);
    while (fbits > 0)  {
        g.push_back(!bit);
        fbits--;
    }
}

Mcell<bool> Bcell::ArithEncoder::finalize() {
    if (finalized) {
        cout << "Trying to finalize() a finalized arithmetic encoder!" << endl;
        exit(EXIT_FAILURE);
    }

    fbits++;

    if (low < FIRST_QTR) {
        bitPlusFollow(0);
    } else {
        bitPlusFollow(1);
    }

    finalized = true;

    Mcell<bool> ret;
    ret = g;
    return ret;
}

Bcell::ArithEncoder::~ArithEncoder() { }

// ARITHDECODER

Bcell::ArithDecoder::ArithDecoder(Bcell *pc) {
    pcell = pc;
    loc = 0;
    value = 0;

    for (unsigned int i = 0; i < CODE_VALUE_BITS; i++)  {
        value = 2 * value + inputBit();
    }

    low = 0;
    high = TOP_VALUE;
}

bool Bcell::ArithDecoder::inputBit() {
    return (loc < pcell->getSize() * 8) ? pcell->getBit(loc++) : 0;
}

int Bcell::ArithDecoder::decodeSymbol() {
    int sym;

    unsigned long int range = high - low + 1;
    unsigned long int cum = ((value - low + 1) * acm[0] - 1) / range;

    for (sym = 0; acm[sym + 1] > cum; sym++);

    high = low + (range * acm[sym]) / acm[0] - 1;
    low  = low + (range * acm[sym + 1]) / acm[0];

    for (;;)  {
        if (high < HALF)  {

        } else if (low >= HALF)  {
            value -= HALF;
            low -= HALF;
            high -= HALF;
        } else if (low >= FIRST_QTR && high < THIRD_QTR) {
            value -= FIRST_QTR;
            low -= FIRST_QTR;
            high -= FIRST_QTR;
        } else {
            break;
        }
        low = 2 * low;
        high = 2 * high + 1;
        value = 2 * value + inputBit();
    }

    acm.update(sym);

    return sym;
}

Bcell::ArithDecoder::~ArithDecoder() { }

// ARITHMODEL

Bcell::ArithModel::ArithModel() : freq(NUM_SYMBOLS, 1), cfreq(NUM_SYMBOLS + 1) {
    for (size_t j = 0; j <= NUM_SYMBOLS; j++) {
        cfreq[j] = NUM_SYMBOLS - j;
    }
}

void Bcell::ArithModel::update(int sym) {
    if (cfreq[0] >= MAX_FREQUENCY) {
        int cum = 0;
        cfreq[NUM_SYMBOLS] = 0;
        for (int i = NUM_SYMBOLS - 1; i >= 0; i--) {
            freq[i] = (freq[i] + 1) / 2;
            cum += freq[i];
            cfreq[i] = cum;
        }
    }

    freq[sym] += INCREMENT_VALUE;
    for (int i = sym; i >= 0; i--) {
        cfreq[i] += INCREMENT_VALUE;
    }
}

const size_t & Bcell::ArithModel::operator[](size_t i) const {
    return cfreq[i];
}

Bcell::ArithModel::~ArithModel() { }
