// bcell_bwt_stree.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// Burrows-Wheeler Transformation functions, Suffix Tree algorithm
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include "bcell.hh"

size_t Bcell::SuffixNode::count = 0;
size_t Bcell::TransitionList::ctor = 0;
size_t Bcell::TransitionList::insert = 0;
size_t Bcell::TransitionList::haschild = 0;
size_t Bcell::TransitionList::doesexist = 0;
size_t Bcell::TransitionList::getfirst = 0;
size_t Bcell::TransitionList::removefirst = 0;
size_t Bcell::TransitionList::find = 0;
size_t Bcell::TransitionList::dtor = 0;

// ********** TRANSITION LIST

void Bcell::TransitionList::displayUsage() {
    cout << "TransitionList constructor calls:             " << ctor << endl;
    cout << "TransitionList insertTransition() calls:      " << insert << endl;
    cout << "TransitionList hasChildren() calls:           " << haschild << endl;
    cout << "TransitionList doesTransitionExist() calls:   " << doesexist << endl;
    cout << "TransitionList getFirstChild() calls:         " << getfirst << endl;
    cout << "TransitionList removeFirstTransition() calls: " << removefirst << endl;
    cout << "TransitionList findTransition() calls:        " << find << endl;
    cout << "TransitionList destructor calls:              " << dtor << endl;
}

Bcell::Transition Bcell::TransitionList::getFirstChild() const {
    getfirst++;

    if (hasChildren()) {
        return lst.front();
    } else {
        cout << "Problem in TransitionList::getFirstChild()" << endl;
        exit(EXIT_FAILURE);
    }
}

void Bcell::TransitionList::insertTransition(const Transition& t) {
    insert++;

    if (!hasChildren()) { // Case 1: List is empty.
        lst.push_front(t);
        return;
    }
    
    if (t.letter < lst.front().letter) { // Case 2: Insert at front of list.
        lst.push_front(t);
        return;    
    }

    for (list<Transition>::iterator it = lst.begin(); it != lst.end(); ++it) {
        if (t.letter < it->letter) { // Case 3: Insert somewhere in middle of list.
            lst.insert(it, t);      // Insert t before it
            return;
        }

        if (t.letter == it->letter) { // Case 4: Edit something in the list.
            *it = t;
            return;
        }
    }

    lst.push_back(t); // Case 5: Insert at end of list.
}

void Bcell::TransitionList::removeFirstTransition() {
    removefirst++;

    if (hasChildren()) {
        lst.pop_front();
    } else {
        cout << "Tried to remove a transition from an empty TransitionList!" << endl;
        exit(EXIT_FAILURE);
    }
}

Bcell::Transition Bcell::TransitionList::findTransition(short a) const {
    find++;

    for (list<Transition>::const_iterator it = lst.begin(); it != lst.end(); ++it) {
        if (it->letter == a) {
            return *it;
        }
    }
    
    cout << "TransitionList::findTransition() could not find the transition." << endl;
    exit(EXIT_FAILURE);
}

bool Bcell::TransitionList::doesTransitionExist(short a) const {
    doesexist++;

    for (list<Transition>::const_iterator it = lst.begin(); it != lst.end(); ++it) {
        if (it->letter == a) {
            return true;
        }
    }

    return false;
}

// ********** SUFFIX NODE

bool Bcell::SuffixNode::haschildren() const {
    return children.hasChildren();
}

// The stack contains internal nodes, never leaf nodes.
void Bcell::SuffixNode::dfs(vector<size_t>& v, SuffixNode * root) {
    stack<pair<SuffixNode *, size_t> > stak;
    stak.push(make_pair(root, (size_t) 0));

    while (!stak.empty()) {
        // stak.top().first points to the node we're processing.
        if (stak.top().first->haschildren()) {
            Transition temp = stak.top().first->children.getFirstChild();
            stak.top().first->children.removeFirstTransition();
            // temp.child points to the SuffixNode below us.
            // Is it a leaf?
            if (temp.child->haschildren()) {
                // There is an internal node below us. Let's process it.
                pair<SuffixNode *, size_t> mypair(temp.child, stak.top().second + temp.right - temp.left + 1);
                stak.push(mypair);
            } else { // There is a leaf below us.
                v.push_back(stak.top().second + temp.right - temp.left + 1); // Update v
                delete temp.child;  // Delete the leaf
            }
        } else {    // We're finished with this internal node.
            delete stak.top().first;    // Delete the internal node.
            stak.pop();
        }
    }
}

// void Bcell::SuffixNode::dfs(list<size_t>& l, size_t x) {
//     if (!haschildren()) {
//         l.push_back(x);
//     } else {
//         while (haschildren()) {
//             Transition temp = children.getFirstChild();
//             temp.child->dfs(l, x + temp.right - temp.left + 1);
//             delete temp.child;
//             children.removeFirstTransition();
//         }
//     }
// }
// 
//     list<size_t> l;
//     Clock dfstimer("DFS Timer");
//     root.dfs(l, 0);
//     dfstimer.print();

// ********** SUFFIX TREE

Bcell::SuffixTree::SuffixTree(const Bcell& b) 
    : root(new SuffixNode),
      text("Suffix Tree Text", b.getSize() + 1),
      infinity(b.getSize()),  
      transformed("BWT Transformed Cell", b.getSize() + 1) {

    // There will be 2 * b.getSize() + 2 SuffixNodes
    // There will be 2 * b.getSize() + 258 TransitionNodes

    for (int j = 1; j <= 257; j++) {
        bottom.children.insertTransition(Transition(sanitizetext(-j), -j, -j, root));
    }
    
    root->link = &bottom;
    
    for (size_t i = 0; i < b.getSize(); i++) {
        text[i] = b[i];
    }

    text[infinity] = 256;           // Sentinel character
    
    // Algorithm 2, steps 4 - 8
    
    {
        Clock ukkonentimer("Ukkonen Timer");
        pair<SuffixNode *, long> temp(root, 0);
        
        for (long i = 0; i < (long) text.getSize(); i++) {
            temp = update(temp, i);
            temp = canonize(temp.first, temp.second, i);
        }
        ukkonentimer.print();
    }
    
    // Now traverse the tree.

    vector<size_t> v;
    v.reserve(b.getSize() + 1);
    Clock dfstimer("DFS Timer");
    SuffixNode::dfs(v, root);
    dfstimer.print();
    cout << "v.size(): " << v.size() << "; b.getSize() + 1: " << b.getSize() + 1 << endl;
            
    const size_t bsize = b.getSize();
    const size_t vsize = v.size();

    Bcell primaryindex;
    Bcell sentinelindex;

    for(size_t i = 0; i < vsize; i++) {
        const size_t vi = v[i];

        if (vi < bsize) {
            transformed[i] = b[bsize - vi];
        } else if (vi == bsize) {
            primaryindex = (unsigned long long) i;
            transformed[i] = b[0];
        } else if (vi == bsize + 1) {               // We must write the sentinel character to transformed
            sentinelindex = (unsigned long long) i; // Record this position for later use
            transformed[i] = 0;                     // Dummy character
        }
    }

    transformed = primaryindex + sentinelindex + transformed;
}

// Ukkonen: mypair.first = s  mypair.second = k
pair<Bcell::SuffixNode *, long> Bcell::SuffixTree::update(pair<SuffixNode *, long> mypair, long i) {
    SuffixNode * oldr = root;
    pair<bool, SuffixNode *> temp = test_and_split(mypair.first, mypair.second, i - 1, sanitizetext(i));
    bool endpoint = temp.first;
    SuffixNode * r = temp.second;
    while (!endpoint) {
        r->children.insertTransition(Transition(sanitizetext(i), i, infinity, new SuffixNode));
        
        if (oldr != root) {
            oldr->link = r;
        }
        oldr = r;
        mypair = canonize(mypair.first->link, mypair.second, i - 1);
        temp = test_and_split(mypair.first, mypair.second, i - 1, sanitizetext(i));
        endpoint = temp.first;
        r = temp.second;    
    }
    if (oldr != root) {
        oldr->link = mypair.first;
    }
    return mypair;
}

pair<bool, Bcell::SuffixNode *> Bcell::SuffixTree::test_and_split(Bcell::SuffixNode * s, long k, long p, short t) {
    if (k <= p) {
        Transition temp = s->children.findTransition(sanitizetext(k));
        long kprime = temp.left;
        long pprime = temp.right;
        SuffixNode * sprime = temp.child;
        
        if (t == sanitizetext(kprime + p - k + 1)) {
            return make_pair(true, s);
        } else {
            SuffixNode * r = new SuffixNode;
            
            s->children.insertTransition(
                Transition(sanitizetext(kprime), kprime, kprime + p - k, r));
            r->children.insertTransition(
                Transition(sanitizetext(kprime + p - k + 1), kprime + p - k + 1, pprime, sprime));
            
            return make_pair(false, r);
        }
    }

    return make_pair(s->children.doesTransitionExist(t), s);
}

// Ukkonen: temp.left = k'  temp.right = p'  temp.child = s'
pair<Bcell::SuffixNode *, long> Bcell::SuffixTree::canonize(Bcell::SuffixNode * s, long k, long p) {
    if (p >= k) {
        Transition temp = s->children.findTransition(sanitizetext(k));
        
        while (temp.right - temp.left <= p - k) {
            k = k + temp.right - temp.left + 1;
            s = temp.child;
            if (k <= p) {
                temp = s->children.findTransition(sanitizetext(k));
            }        
        }
    }

    return make_pair(s, k);
}

short Bcell::SuffixTree::sanitizetext(long x) {
    if (x >= 0) {
        return text[x];
    } else if (x >= -257) {
        return -x - 1;
    } else {
        cout << "Dereferenced huge negative number in Suffix Tree functions!" << endl;
        exit(EXIT_FAILURE);
    }
}

Mcell<unsigned char> Bcell::SuffixTree::getTransformed() const {
    return transformed;
}

// ********** TRADITIONAL WRAPPER FUNCTIONS

Bcell& Bcell::bwtCompressStree(bool verbose) {
    Bcell sig(BWTZIP_STREE_SIG);

    Bcell id("Compression Method ID", 1);
    id[0] = 255;
    
    if (size <= MIN_LENGTH_TO_COMPRESS) {
        id[0] = 0;
        *this = sig + id + *this;
        return *this;
    }

    Mcell<long double> timings("Timing Information", 4);

    clockStart(verbose);

    bwtStree();     timings[0] = clockReport();
    mtf2();         timings[1] = clockReport();
    zleWheeler();   timings[2] = clockReport();
    arith();        timings[3] = clockReport();

    if (verbose) {
        cout << "BWT Stree:   " << timings[0] << endl;
        cout << "MTF-2:       " << timings[1] << endl;
        cout << "Wheeler ZLE: " << timings[2] << endl;
        cout << "Arith:       " << timings[3] << endl;
    }

    id[0] = 1;
    *this = sig + id + *this;

    return *this;
}

Bcell& Bcell::bwtDecompressStree(bool verbose) {
    if (size < 8 || readULL(0) != BWTZIP_STREE_SIG) {
        cout << "Not compressed with bwtzip/stree!" << endl;
        exit(EXIT_FAILURE);
    }

    unsigned char id = (*this)[8];
    truncate(9, size);
    
    if (id == 0) { // Not compressed
        return *this;
    }

    clockStart(verbose);
    Mcell<long double> timings("Timing Information", 4);
    
    unarith();      timings[0] = clockReport();
    unzleWheeler(); timings[1] = clockReport();
    unmtf2();       timings[2] = clockReport();
    unbwtStree();   timings[3] = clockReport();
    
    if (verbose) {
        cout << "Unarith:       " << timings[0] << endl;
        cout << "UnWheeler ZLE: " << timings[1] << endl;
        cout << "UnMTF-2:       " << timings[2] << endl;
        cout << "UnBWT Stree:   " << timings[3] << endl;
    }

    return *this;
}

// Burrows-Wheeler Transform our cell
Bcell& Bcell::bwtStree() {
    if (size < 2) {
        cout << "ERROR: Bcell of length less than 2 passed to bwtStree()." << endl;
        exit(EXIT_FAILURE);
    }

    SuffixNode::displayInstances();
    {
        SuffixTree suffixtree(*this);
        *this = suffixtree.getTransformed();
    }
    SuffixNode::displayInstances();
    TransitionList::displayUsage();

    return *this;
}

// Reverse Burrows-Wheeler Transform our cell using the Fenwick algorithm
Bcell& Bcell::unbwtStree() {
    size_t index = readULL(0);                   // Read primary index
    size_t sentinelindex = readULL(8);           // Read sentinel index
    truncate(16, size);                          // Remove encoded indices
    Bcell dcell("UnBWT Decoded Cell", size);     // Holds original data

    Mcell<size_t> freqs("Frequencies", 257);     // Fenwick's K

    freqs.fill(0);

    for (size_t st = 0; st < size; st++) {
        if (st == sentinelindex) {
            freqs[256]++;
        } else {
            freqs[(*this)[st]]++;
        }
    }

    Mcell<size_t> links("Links", size);         // Fenwick's L
    Mcell<size_t> mapping("Mapping", 257);      // Fenwick's M

    mapping[0] = 0;                             // Build M
    for (int j = 1; j < 257; j++) {
        mapping[j] = mapping[j - 1] + freqs[j - 1];
    }
   
    for (size_t i = 0; i < size; i++) {         // Build L
        if (i == sentinelindex) {
            links[i] = mapping[256]++;
        } else {
            links[i] = mapping[(*this)[i]]++;
        }
    }

    for (long long int i = size - 1; i >= 0; i--) {         // Perform Reverse-BWT
        index = links[index];
        dcell[i] = (*this)[index];        
    }

    *this = dcell.truncate(0, size - 1);    // Remove sentinel character
    return *this;
}
