// bcell_huff.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// Huffman functions
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include "bcell.hh"

// Compress our cell using Huffman compression and return our cell
Bcell& Bcell::huff() {
    const Mcell<size_t>       freqs       = getFrequencies();
    const Bcell               codelengths = HuffmanCodeLengths();
    const Mcell<Gcell<bool> > codes       = codelengths.HuffmanCodes();

    Gcell<bool> temp;

    for (size_t i = 0; i < size; i++) {  // For each byte in our cell
        temp += codes[(*this)[i]];
    }

    const Bcell sizeCell((unsigned long long) size);
    const Bcell bits = temp.getMcell();
    *this = sizeCell + codelengths + bits;
    return *this;
}

// Decompress our cell using Huffman compression and return our cell
Bcell& Bcell::puff() {

    Bcell codelengths("puff() Codelengths", 256);
    codelengths.transcopy(256, 8, *this, 0);

    const Mcell<Gcell<bool> > codes = codelengths.HuffmanCodes();
    Bcell ret("Decompressed Huffman cell", readULL(0));

    Gcell<bool> currentcode;
    size_t n = 0;   // Location in decompressed cell being written to

    for (size_t i = 2112; i < size * 8; i++) {
        currentcode.appendData(getBit(i));

        for (int j = 0; j < 256; j++) {
            if (currentcode == codes[j]) {
                ret[n++] = j;
                currentcode.clear();
                break;
            }
        }
    }

    *this = ret;
    return *this;
}

// Given codelengths, calculate the canonical Huffman codes
// Returns an Mcell of length 256
Mcell<Gcell<bool> > Bcell::HuffmanCodes() const {

    Mcell<Gcell<bool> > ret("Huffman Codes Mcell<Gcell<bool> >", 256);

    if (size != 256) {
        cout << "ERROR: HuffmanCodes() reports length not 256." << endl;
        exit(EXIT_FAILURE);
    }

    Bcell codes("Huffman codes", 256); // Stores the lower 8 bits, right-aligned, of each code
    Mcell<size_t> codelengthfreqs = getFrequencies();

    // Calculate right-aligned start positions

    Bcell starts("Huffman code start positions", 256);

    unsigned char s = 0;

    for (int i = 255; i >= 0; i--) {
        starts[i] = s;
        s += codelengthfreqs[i];
        s >>= 1;
    }

    // Now fill in codes

    for (int i = 255; i >= 0; i--) {
        unsigned char c = starts[i];

        // Fill in codes with codelength i
        for (int j = 0; j < 256; j++) {
            if ((*this)[j] == i) {
                codes[j] = c++;
            }
        }
    }

    for (int i = 0; i < 256; i++) {
        unsigned char ourcode = codes[i];       // Lower 8 bits, right-aligned
        unsigned char ourlength = (*this)[i];
        for (int j = 0; j < ourlength; j++) {
            ret[i].prependData(ourcode & 1);
            ourcode >>= 1;
        }
    }

    return ret;
}

// Calculate Huffman codelengths
// Returns a Bcell of length 256
Bcell Bcell::HuffmanCodeLengths() const {
    Mcell<size_t> freq = getFrequencies();
    Bcell codelengths("Huffman Codelengths", 256);

    Mcell<Tree> ta("Tree Array", 511);
    Mcell<bool> ba("bool Array", 511);

    ba.fill(false);

    // Initialize trees. ba stores which trees are currently active.

    for (int i = 0; i < 256; i++) {
        ta[i].set(freq[i], i);
        ba[i] = true;
    }

    // The highest possible frequency of a tree is size. So size + 1 is
    // larger than any possible frequency.

    for (int i = 256; i < 511; i++) {
        // Fill in the next intermediate tree, which is ta[i].

        // Find the lowest-frequency tree.
        int l1 = -1;
        size_t f1 = size + 1;

        for (int j = 0; j < i; j++) {
            if (ba[j] && ta[j].freq < f1) {
                l1 = j;
                f1 = ta[j].freq;
            }
        }

        ba[l1] = false;

        // Find the second-lowest-frequency tree.
        int l2 = -1;
        size_t f2 = size + 1;

        for (int j = 0; j < i; j++) {
            if (ba[j] && ta[j].freq < f2) {
                l2 = j;
                f2 = ta[j].freq;
            }
        }

        ba[l2] = false;

        ta[i].set(ta[l1], ta[l2]);
        ba[i] = true;
    }

    codelengths.treefill(ta[510]);

    return codelengths;
}

// Fill our cell (of length 256) with the codelengths of a tree
void Bcell::treefill(Tree & t, unsigned char h) {
    if (size != 256) {
        cout << "ERROR: treefill() reports length not 256." << endl;
        exit(EXIT_FAILURE);
    }

    int byte = t.byte;

    if ((0 <= byte) && (byte <= 255)) {
        (*this)[byte] = h;
        return;
    }

    // We are at an intermediate node

    Tree * left = t.left;
    Tree * right = t.right;

    if ((byte != -1) || (left == NULL) || (right == NULL)) {
        cout << "ERROR: Problem in Bcell::treefill()" << endl;
        exit(EXIT_FAILURE);
    }

    treefill(*left, h + 1);
    treefill(*right, h + 1);
}

// Constructor: Makes an empty tree
Bcell::Tree::Tree() {
    set(0, -1);
}

// Make our tree into a one-node tree
void Bcell::Tree::set(size_t f, int b) {
    freq = f;
    byte = b;
    left = NULL;
    right = NULL;
    up = NULL;
}

// Creates two trees from one
void Bcell::Tree::set(Tree & l, Tree & r) {
    l.up = this;
    r.up = this;
    freq = l.freq + r.freq;
    byte = -1;
    left = &l;
    right = &r;
    up = NULL;
}

// Destructor
Bcell::Tree::~Tree() {
    // Goes Nowhere Does Nothing
}

// Find where a given byte is
const Bcell::Tree * Bcell::Tree::findByte(int b) const {
    if (byte == b) {
        return this;
    }
    if ((left != NULL) && (left->findByte(b) != NULL)) {
        return left->findByte(b);
    }
    if ((right != NULL) && (right->findByte(b) != NULL)) {
        return right->findByte(b);
    }
 
    return NULL;
}

// Get the level at which a given byte is at
int Bcell::Tree::getHeight(int b) const {
    const Tree * t = findByte(b);
    int h = 0;

    if (t == NULL) {
        cout << "Could not find a byte in the tree!" << endl;
        exit(EXIT_FAILURE);
    }

    while (t->up != NULL) {
        h++;
        t = t->up;
    }

    return h;
}
