// bcell_srle.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// Switched-RLE functions
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include "bcell.hh"

// Phil Norman's Switched-RLE algorithm (srle)
// The compressor and decompressor can be in one of two states,
// LITERAL and FILL, and they start in LITERAL. The state means
// what type of run is currently being described. Both types of
// runs are prefixed by a single header byte, which describes
// how long the run is. A LITERAL header is then followed by
// that many bytes in the clear. A FILL header stands alone
// and represents a number of identical bytes. The byte used in
// the fill is the last literal byte which was output.
// The states are not directly encoded in the output. Instead,
// the length of the runs signals what state the next run will
// be. If the previous header was not 255, then the next state
// will be different. If the previous header was 255, then
// the next state will stay the same.
// To encode a run of exactly 255 elements, one header of
// 255 followed by one header of 0 is used.

// Stephan T. Lavavej's Switched-RLE-0 algorithm (srle0)
// An optimization for BWT=>MTF data.
// SRLE-0 works exactly like SRLE, except that FILL runs are
// always assumed to be of byte 0x00. Literal streams do not
// have to end in 0x00. Indeed, literal streams should never
// end with a 0x00 (unless it is the end of the data) and they
// should never contain two 0x00s in a row. A single 0x00 is
// encoded literally.

// Stephan T. Lavavej's Switched-RLE-0-A algorithm (srle0a)
// A further optimization for BWT=>MTF data.
// SRLE-0-A is is exactly like SRLE-0, except that longer FILL runs
// can be represented. FILL runs of length 0 to 127 inclusive
// are represented normally. FILL runs of length 128 to 32767
// inclusive are represented by a 2-byte header (big-endian), whose
// highest bit is on. The rest of the algorithm is appropriately
// modified (e.g. testing whether to switch states, a switch happens
// after a LITERAL run of 255 or a FILL run of 32767).

// Switched-RLE encode our cell
Bcell& Bcell::srle() {

    Gcell<unsigned char> output         ; 
    Gcell<unsigned char> buf            ;
    bool                 literal  = true;
    int                  count    = 0   ;
    size_t               n        = 0   ;
    unsigned char        fillbyte = 0   ;

    while (n < size) {
        // We are looking at the start of a new run
        if (literal) { // LITERAL
            while (n < size && count < 255) {
                fillbyte = (*this)[n];
                buf.appendData((*this)[n]);
                n++;
                count++;
                
                if (n != size && (*this)[n - 1] == (*this)[n]) {
                    break;
                }
            }
        } else { // FILL
            // We are looking for up to 255 identical bytes.
            while (n < size && count < 255) {
                if ((*this)[n] == fillbyte) {
                    n++;
                    count++;
                } else {
                    break;
                }
            }
        }
        
        // Run is over. Header byte must be count.
        
        output.appendData(count);

        if (literal) { // LITERAL
            output.destructiveAppend(buf);
        }
        
        if (count != 255) {
            literal = !literal;
        }
        
        count = 0;
    }
    
    *this = output.getMcell();
    return *this;
}

// Switched-RLE decode our cell
Bcell& Bcell::unsrle() {
    Gcell<unsigned char> output;
    bool literal = true;
    size_t n = 0;
    unsigned char fillbyte = 0;

    while (n < size) {
        unsigned char header = (*this)[n++];        

        if (literal) { // LITERAL
            for (int i = 0; i < header; i++) {
                fillbyte = (*this)[n++];
                output.appendData(fillbyte);
            }
        } else { // FILL
            for (int i = 0; i < header; i++) {
                output.appendData(fillbyte);
            }
        }
        
        if (header != 255) {
            literal = !literal;
        }
    }

    *this = output.getMcell();
    return *this;
}

// Switched-RLE-0 encode our cell
Bcell& Bcell::srle0() {

    Gcell<unsigned char> output;
    Gcell<unsigned char> buf;
    bool literal = true;
    int count = 0;
    size_t n = 0;

    while (n < size) {
        // We are looking at the start of a new run
        if (literal) { // LITERAL
            while (n < size && count < 255) {
                if (n == size - 1 || (*this)[n] != 0 || (*this)[n + 1] != 0) {
                    buf.appendData((*this)[n]);
                    n++;
                    count++;
                } else { // We are at the beginning of a run of zeroes
                    break;
                }            
            }
        } else { // FILL
            // We are looking for up to 255 zeroes.
            while (n < size && count < 255) {
                if ((*this)[n] == 0) {
                    n++;
                    count++;
                } else {
                    break;
                }
            }
        }
        
        // Run is over. Header byte must be count.
        
        output.appendData(count);

        if (literal) { // LITERAL
            output.destructiveAppend(buf);
        }
        
        if (count != 255) {
            literal = !literal;
        }
        
        count = 0;
    }
    
    *this = output.getMcell();
    return *this;
}

// Switched-RLE-0 decode our cell
Bcell& Bcell::unsrle0() {
    Gcell<unsigned char> output;
    bool literal = true;
    size_t n = 0;

    while (n < size) {
        unsigned char header = (*this)[n++];        

        if (literal) { // LITERAL
            for (int i = 0; i < header; i++) {
                output.appendData((*this)[n++]);
            }
        } else { // FILL
            for (int i = 0; i < header; i++) {
                output.appendData(0);
            }
        }
        
        if (header != 255) {
            literal = !literal;
        }
    }

    *this = output.getMcell();
    return *this;
}

// Switched-RLE-0-A encode our cell
Bcell& Bcell::srle0a() {

    Gcell<unsigned char> output;
    Gcell<unsigned char> buf;
    bool literal = true;
    int count = 0;
    size_t n = 0;

    while (n < size) {
        // We are looking at the start of a new run
        if (literal) { // LITERAL
            while (n < size && count < 255) {
                if (n == size - 1 || (*this)[n] != 0 || (*this)[n + 1] != 0) {
                    buf.appendData((*this)[n]);
                    n++;
                    count++;
                } else { // We are at the beginning of a run of zeroes
                    break;
                }            
            }
        } else { // FILL
            // We are looking for up to 32767 zeroes.
            while (n < size && count < 32767) {
                if ((*this)[n] == 0) {
                    n++;
                    count++;
                } else {
                    break;
                }
            }
        }
        
        // Run is over. Header byte must be count.
                
        if (literal) {
            output.appendData(count);
        } else {
            if (count < 128) {
                output.appendData(count); // High bit 0
            } else {
            	output.appendData(((count >> 8) & 0xFF) | 0x80);
                output.appendData(count & 0xFF);
            }
        }

        if (literal) { // LITERAL
            output.destructiveAppend(buf);
        }
        
        if ((literal && count != 255) || (!literal && count != 32767)) {
            literal = !literal;
        }
        
        count = 0;
    }
    
    *this = output.getMcell();
    return *this;
}

// Switched-RLE-0-A decode our cell
Bcell& Bcell::unsrle0a() {
    Gcell<unsigned char> output;
    bool literal = true;
    size_t n = 0;

    while (n < size) {
        unsigned int header = (*this)[n++];

        if (literal) { // LITERAL
            for (unsigned int i = 0; i < header; i++) {
                output.appendData((*this)[n++]);
            }
        } else { // FILL
        	if (header & 0x80) {
        		header &= 0x7F;
        		header <<= 8;
        		header |= (*this)[n++];
        	}
        
            for (unsigned int i = 0; i < header; i++) {
                output.appendData(0);
            }
        }

        if ((literal && header != 255) || (!literal && header != 32767)) {
            literal = !literal;
        }

    }

    *this = output.getMcell();
    return *this;
}
