// bcell_zlib.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// Zlib functions
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include "bcell.hh"

#ifdef ZIP

// Compress our cell as much as possible using zlib, returning our cell
Bcell& Bcell::compress() {

    Bcell t1 = compressStrategy(Z_DEFAULT_STRATEGY);
    Bcell t2 = compressStrategy(Z_FILTERED);
    Bcell t3 = compressStrategy(Z_HUFFMAN_ONLY);

    if (t1.size <= t2.size && t1.size <= t3.size) {
        *this = t1;
    } else if (t2.size <= t1.size && t2.size <= t3.size) {
        *this = t2;
    } else {
        *this = t3;
    }

    return *this;
}

// This has to play with "bare" memory, unfortunately.
Bcell Bcell::compressStrategy(int num_strategy) const {

    unsigned long int sourcelen, destlen;
    unsigned char *source, *dest;

    sourcelen = size;
    destlen = sourcelen + (sourcelen >> 9) + 12;    /* Slightly more than required by zlib */

    source = new unsigned char[sourcelen];
    dest = new unsigned char[destlen];

    for (size_t st = 0; st < sourcelen; st++) {
        source[st] = (*this)[st];
    }

    int error = myCompress(dest, &destlen, source, sourcelen, num_strategy);

    // Upon exit, destlen is the actual size of the compressed buffer.

    if (error != Z_OK) {
        cout << "ERROR: zlib reported error in compressStrategy()." << endl;
        exit(EXIT_FAILURE);
    }

    Bcell compressed("Compressed Bcell", destlen);

    for (size_t st = 0; st < destlen; st++) {
        compressed[st] = dest[st];
    }

    delete[] source;
    delete[] dest;

    return compressed;
}

// This has to play with "bare" memory, unfortunately.
Bcell& Bcell::decompress(size_t originalsize) {

    unsigned long int sourcelen, destlen;
    unsigned char *source, *dest;

    sourcelen = size;
    destlen = originalsize;

    source = new unsigned char[sourcelen];
    dest = new unsigned char [destlen];

    for (size_t st = 0; st < sourcelen; st++) {
        source[st] = (*this)[st];
    }

    int error = uncompress(dest, &destlen, source, sourcelen); 

    if (error != Z_OK) {
        cout << "ERROR: zlib reported error in decompress()." << endl;
        exit(EXIT_FAILURE);
    }

    if (destlen != originalsize) {
        cout << "ERROR: Decompressed data takes an unexpected amount of memory." << endl;
        exit(EXIT_FAILURE);
    }

    Bcell uncompressed("Uncompressed Bcell", destlen);

    for (size_t st = 0; st < destlen; st++) {
        uncompressed[st] = dest[st];
    }

    delete[] source;
    delete[] dest;

    *this = uncompressed;
    return *this;
}

// Modified compress2() function from zlib that calls deflateinit2().
// It has the power of init2 with the convenience of compress.
int Bcell::myCompress(
    unsigned char *dest, unsigned long int *destlen,
    const unsigned char *source, unsigned long int sourcelen,
    int num_strategy) {

    z_stream the_stream;
    int err;

    the_stream.next_in = (unsigned char *) source;
    the_stream.avail_in = (unsigned int) sourcelen;
    the_stream.next_out = dest;
    the_stream.avail_out = (unsigned int) *destlen;

    if ((unsigned long) the_stream.avail_out != *destlen) {
        return Z_BUF_ERROR;
    }

    the_stream.zalloc = (alloc_func)0;
    the_stream.zfree = (free_func)0;
    the_stream.opaque = (voidpf)0;

    // level 9, windowBits 15, memLevel 9
    err = deflateInit2(&the_stream, 9, Z_DEFLATED, 15, 9, num_strategy);

    if (err != Z_OK) {
        return err;
    }

    if ((err = deflate(&the_stream, Z_FINISH)) != Z_STREAM_END) {
        deflateEnd(&the_stream);
        return ((err == Z_OK) ? Z_BUF_ERROR : err);
    }

    *destlen = the_stream.total_out;
    return deflateEnd(&the_stream);
}

#endif // End of: #ifdef ZIP
